//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "LaneChangeManeuverBase.h"

Define_Module(LaneChangeManeuverBase);

void LaneChangeManeuverBase::initialize(int stage)
{
    BaseScenario::initialize(stage);
    if (stage == 0) {
        //get pointer to application
        appl = FindModule<BaseApp*>::findSubModule(getParentModule());
    }
    if (stage == 1) {
        platooningVType = par("platooningVType").stdstringValue();
        traciVehicle->setFixedLane(1, false);
        traciVehicle->setSpeedMode(0);
        if (positionHelper->isLeader()) {
            for (int i = 1; i < positionHelper->getPlatoonSize(); i++) {
                std::stringstream ss;
                ss << platooningVType << "." << positionHelper->getMemberId(i);
                traciVehicle->addPlatoonMember(ss.str(), i);
            }
            traciVehicle->setCruiseControlDesiredSpeed(mobility->getSpeed());
        } else {
            std::stringstream ssl, ss;
            ssl << platooningVType << "." << positionHelper->getLeaderId();
            ss << platooningVType << "." << positionHelper->getFrontId();
            //traciVehicle->setCruiseControlDesiredSpeed(mobility->getSpeed() + 10);
        }

        BaseProtocol *protocol = FindModule<BaseProtocol*>::findSubModule(getParentModule());

        //connect maneuver application to protocol
        protocol->registerApplication(MANEUVER_TYPE, gate("lowerLayerIn"), gate("lowerLayerOut"), gate("lowerControlIn"),gate("lowerControlOut"));
        //connect beaconing application to protocol
        protocol->registerApplication(BEACON_TYPE, gate("lowerLayerIn"), gate("lowerLayerOut"), gate("lowerControlIn"),gate("lowerControlOut"));
    }
}
/**
 * Returns the current lane index of the freeway
 */
int LaneChangeManeuverBase::getLaneId() {
    return traciVehicle->getLaneIndex();
}

/**
 * Returns a boolean value indication whether this vehicle is a the platoon leader or not
 */
bool LaneChangeManeuverBase::isLeader() {
    return positionHelper->isLeader();
}
/**
 * Returns the id of the vehicle in front of this vehicle
 */
int LaneChangeManeuverBase::getFrontId() {
    return positionHelper->getFrontId();
}

/**
 * Retuns the id of the vehicle in the back of this vehicle
 */
int LaneChangeManeuverBase::getBackId() {
    return positionHelper->getBackId();
}
/**
 * Returns the id of the current car in the platoon.
 * 0 represents the leader
 */
int LaneChangeManeuverBase::getMyPlatoonId() {
    return positionHelper->getId();
}
/*
 * Returns the number of vehicles in this platoon
 */
int LaneChangeManeuverBase::getPlatoonSize() {
    return positionHelper->getPlatoonSize();
}

/*
 * Returns the length of the current vehicle
 */
double LaneChangeManeuverBase::getVehicleLength() {
    return traciVehicle->getLength();
}
/**
 * Sets the speed of the leader. Since the leader is handling a platoon, the acceleration / deceleration is amplified by all follower
 */
void LaneChangeManeuverBase::setPlatoonSpeed(double desiredSpeed) {
    if(!isLeader()) { //Only the leader is allowed to start maneuvers
        endSimulation();
    }
    traciVehicle->setCruiseControlDesiredSpeed(desiredSpeed);
}

/**
 * Changes the lane without considering other traffic -> If the new lane is blocked by vehicles, there will be a crash
 * laneIndex: The new lane
 */
void LaneChangeManeuverBase::changeLane(int laneIndex)
{
     traciVehicle->setFixedLane(laneIndex, false); //
}

/**
 * msg: The cPacket to send
 * destination: Vehicle destination - use -1 for broadcast
 */
void LaneChangeManeuverBase::sendManeuverMessage(cPacket *msg, int destination) {

    if(!isLeader()) { //Only the leader is allowed to start maneuvers
        endSimulation();
    }
    UnicastMessage *unicast = new UnicastMessage("", MANEUVER_TYPE);
    unicast->setDestination(destination);
    unicast->setChannel(Channels::CCH);
    unicast->encapsulate(msg);
    sendDown(unicast);
}
/**
 *  Returns the current position as X and Y values
 */
Coord LaneChangeManeuverBase::getMyPosition() {
    return  mobility->getCurrentPosition();//mobility->getManager()->omnet2traci(mobility->getCurrentPosition());
}
